
// fn       port        16F1503 pin
// I2C SCK  C0          10
// I2C SDA  C1          9
// Red      C5/PWM1     5
// Blue     C3/PWM2     7
// Green    A2/PWM3     11
// X        C4          6

#define I2C_ADDR        0x44

#pragma config FOSC = INTOSC    // internal clock
#pragma config WDTE = OFF       // watchdog timer disabled
#pragma config PWRTE = ON       // power-up timer enabled
#pragma config MCLRE = OFF      // MCLR/Vpp pin is digital input
#pragma config BOREN = OFF      // brown-out reset disabled
#pragma config WRT = OFF        // write Protection off
#pragma config LVP = OFF        // low voltage programmer disabled

#include <xc.h>
#include <limits.h>

#define CMD_LEADIN      0xfe
#define CMD_NOOP        0x00 // no params, do nothing
#define CMD_ALLOFF      0x01 // no params, set all outputs off
#define CMD_SETRGB      0x02 // params are r, g, b duty cycles
#define CMD_BALANCE     0x08 // params are r, g, b balances
#define CMD_SETX        0x04 // param1 is x output on if nonzero
#define CMD_XPULSE      0x05 // param1 is x on-time

#define COMMANDSIZE     7
union
{
    unsigned char Buffer[COMMANDSIZE];
    struct
    {
        unsigned char Leadin;
        unsigned char BlockSize;    // smbus block write
        unsigned char Command;
        unsigned char Param1;
        unsigned char Param2;
        unsigned char Param3;
        unsigned char Checksum;     // XOR of leadin, command, params
    } B;
} Cmd;
unsigned char CmdBuffer_i;

// color balance
unsigned int R_scale;
unsigned int G_scale;
unsigned int B_scale;

void set_rgbscale(unsigned char r, unsigned char g, unsigned char b)
{
    R_scale = (unsigned int)r + 2;
    G_scale = (unsigned int)g + 2;
    B_scale = (unsigned int)b + 2;
}

void set_rgb(unsigned char r, unsigned char g, unsigned char b)
{
    union
    {
        unsigned int _int;
        struct
        {
            unsigned char low;
            unsigned char high;
        } _char;
    } c;

    c._int = R_scale * r;
    PWM1DCH = c._char.high;
    PWM1DCL = c._char.low;

    c._int = G_scale * g;
    PWM3DCH = c._char.high;
    PWM3DCL = c._char.low;

    c._int = B_scale * b;
    PWM2DCH = c._char.high;
    PWM2DCL = c._char.low;
}

inline void set_x(unsigned char x)
{
    PORTCbits.RC4 = x ? 1 : 0;
}

void exec_command(void)
{
    unsigned char checksum =
        Cmd.B.Leadin ^
        Cmd.B.Command ^
        Cmd.B.Param1 ^
        Cmd.B.Param2 ^
        Cmd.B.Param3;
    if (checksum != Cmd.B.Checksum)
    {
        return;
    }

    switch (Cmd.B.Command)
    {
        case CMD_SETRGB:
            set_rgb(Cmd.B.Param1, Cmd.B.Param2, Cmd.B.Param3);
            break;

        case CMD_BALANCE:
            set_rgbscale(Cmd.B.Param1, Cmd.B.Param2, Cmd.B.Param3);
            break;

        case CMD_SETX:
            set_x(Cmd.B.Param1);
            break;

        case CMD_XPULSE:
            TMR0 = UCHAR_MAX - Cmd.B.Param1;
            INTCONbits.TMR0IE = 1;
            set_x(1);
            break;

        case CMD_ALLOFF:
            set_rgb(0, 0, 0);
            set_x(0);
            break;

        case CMD_NOOP:
        default:
            break;
    }
}

inline void push_cmd_byte(unsigned char cmdbyte)
{
    if (CmdBuffer_i == 0 && cmdbyte != CMD_LEADIN)
    {
        return;
    }

    if (CmdBuffer_i == 1 && cmdbyte != COMMANDSIZE - 2)
    {
        CmdBuffer_i = 0;
        return;
    }

    Cmd.Buffer[CmdBuffer_i++] = cmdbyte;
    if (CmdBuffer_i == COMMANDSIZE)
    {
        exec_command();
        CmdBuffer_i = 0;
    }
}

void interrupt isr(void)
{
    if (SSP1IF) // I2C
    {
        SSP1IF = 0;
        unsigned char value = SSP1BUF; // BF cleared
        unsigned char stat = SSP1STAT & 0b00100100; // mask to D/!A, R/!W
        if (stat == 0b00100000) // data received
        {
            push_cmd_byte(value);
        }
        else if (stat == 0b00000000) // address received
        {
            CmdBuffer_i = 0;
        }

        SSP1CON1bits.CKP = 1; // release clock
    }

    if (TMR0IF) // timer 0
    {
        INTCONbits.TMR0IE = 0;
        TMR0IF = 0;
        set_x(0);
    }
}

void quickdelay(char d)
{
    for (char i = 0; i < d; i++)
        for (char j = 0; j < 100; j++);
}

void test()
{
    for (char r = 0; r < 250; r++)
    {
        set_rgb(r, 0, 0);
        quickdelay(5);
    }

    for (char g = 0; g < 250; g++)
    {
        set_rgb(0, g, 0);
        quickdelay(5);
    }

    for (char b = 0; b < 250; b++)
    {
        set_rgb(0, 0, b);
        quickdelay(5);
    }

    set_rgb(0, 0, 0);

    set_x(1);
    quickdelay(100);
    set_x(0);
}

void main(void)
{
    OSCCON = 0b01111010;    // internal osc, 16MHz
    while (!HFIOFS);        // wait for osc stable

    CmdBuffer_i = 0;
    set_rgbscale(255, 255, 255);

    // PWM init
    PR2 = 0xff;
    T2CON = 0b00000110;         // timer2 on, prescaler = 16 (0.97KHz PWM)
    set_rgb(0, 0, 0);
    PWM1CON = 0b11000000;       // PWM enable, output enable, active high
    PWM2CON = 0b11000000;
    PWM3CON = 0b11000000;

    set_x(0);
    ANSELA = 0b11111011;        // A2 digital I/O
    ANSELC = 0b11110100;        // C0, C1, C3 digital I/O
    TRISA = 0b11111011;         // A2 output
    TRISC = 0b11000111;         // C0, C1 input; C3, C4, C5 output

    // timer 0 init
    OPTION_REG = 0b11010111;    // clock from fosc/4
                                // prescaler on timer 0
                                // prescaler 1:256

    // I2C init
    SSP1CON1 = 0b00110110;      // serial port enabled
                                // release clock
                                // I2C slave, 7 bit addr
    SSP1CON2 = 0b00000001;      // general call ignored
                                // clock stretching enabled
    SSP1STAT = 0b11000000;      // slew control disabled (standard rate)
                                // enable input logic, smbus compliant
    SSP1MSK = 0xff;
    SSP1ADD = I2C_ADDR << 1;

    test();

    PIE1 = 0b00001000;          // MSSP interrupt enabled
    INTCON = 0b11000000;        // peripheral and global interrupt enabled
                                // timer 0 interrupt disabled

    while (1);
}
